/*
 * file StpLocation.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.StpLocation
 *
 * (C) Copyright IBM Corporation 2004, 2009.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp;

import java.io.File;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.util.EnumSet;
import java.util.Hashtable;
import java.util.Map;

import javax.wvcm.Location;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;

import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.annotation.RpcEnum;

/**
 * An extension of the javax.wvcm Location interface that provides a
 * programmatic representation for the location of a resource.
 * 
 * <p>
 * An StpLocation instance represents a location specification that has been
 * parsed into its various component fields. A number of different formats or
 * <i>schemes</i> are used to express the location of various resources as a
 * string. These schemes consist of one or more of the following fields:
 * <i>domain</i>, <i>repository name</i>, <i>namespace</i>, and <i>object
 * name</i>. It is the namespace field that determines the scheme being used.
 * <p>
 * Locations are hierarchical, with the domain field specifying the top level of
 * the hierarchy. Within a domain, resources are partitioned into repositories.
 * Within a repository, resources are first partitioned into namespaces, and
 * then uniquely identified by segmented pathnames within that namespace.
 * <p>
 * Each scheme requires certain of the above fields to be specified. If required
 * fields are not present, the StpLocation object will have a non-OK Status.
 * Individual field values can be queried to determine which fields aren't
 * present.
 * <p>
 * To use this StpLocation to construct a proxy, its Status must be OK.
 * <p>
 * When a proxy is constructed, a new StpLocation may need to be constructed, so
 * clients must not assume that the object returned by Resource.location()
 * or any of the StpLocation factory methods defined in StpProvider is
 * the same object passed to the proxy factory that created the proxy.
 * <p>
 * The preferred scheme for specifying an object is the <i>object selector
 * scheme</i>, which has the following structure
 * <p> [[<i>&lt;domain></i>.][<i>&lt;namespace></i>]:] [<i>&lt;object-name></i>][&#64;[<i>&lt;repository-name></i>]]
 * <p>
 * The <i>&lt;object-name></i>, and <i>&lt;repository-name></i> fields are
 * segmented names where the segments are separated by '/'s or '\'s. The
 * permitted <i>&lt;namespace></i>s are defined by
 * {@link StpLocation.Namespace} and the permitted <i>&lt;domain></i>s are
 * defined by {@link StpProvider.Domain}.
 * <p>
 * The character '&#64;' is reserved for use as the repository field delimiter
 * as defined above. If it is to be part of the name field, it must be escaped
 * by preceding it with a percent sign '%'. Similarly, the characters '/' and
 * '\' are reserved in both the name and repository fields to be used as
 * pathname segment separators. To use them as part of a segment they, too, must
 * be escaped using a percent sign. Use two percent signs, '%%', to include a
 * percent sign in the name or repository field. Note that escaped characters
 * within a field are <i>not</i> unescaped when parsed into a StpLocation.
 * Utility methods are provided by the StpProvider class for unescaping the
 * fields, should a client need the unadulterated image.
 * <p>
 * Some resources can also be referenced directly or indirectly by an absolute or
 * relative file system pathname. As an StpLocation image, such
 * representations are called <i>path-scheme</i> locations, which have the
 * following structure
 * <p> [[<i>&lt;domain></i>.][<i>&lt;namespace></i>]:] [<i>&lt;path-name></i>]
 * <p>
 * The path-scheme locations are further categorized by their namespace as
 * indicated in this enumeration...
 * <ul>
 * <li>{@linkplain #isOk() <i>invalid</i>} namespaces: NONE, DEFAULT, INVALID,
 * <li>{@linkplain #isFilePathScheme() <i>file path scheme</i>} namespaces:
 * PNAME, PNAME_IMPLIED, FILE,
 * <li>{@linkplain #isRepositoryPathScheme() <i>repository path scheme</i>}
 * namespaces: VOB, VIEW_UUID, REPLICA_UUID, PROJ_DB, USER_DB, DB_SET, SERVER,
 * <li>WORKSPACE,
 * <li>{@linkplain #isUrlPathScheme() <i>URL path scheme</i>} namespaces: FILE,
 * HTTP, HTTPS,
 * </ul>
 * This interface defines a predicate for each of these categories. The reader
 * is referred to the documentation for those predicates for more information on
 * the formation and meaning of each type of path-scheme location. Note that if
 * a namespace is <i>not</i> one of the above path-scheme namespaces, it is an
 * {@linkplain #isObjectSelectorScheme() <i>object selector scheme</i>}
 * namespace.
 * <p>
 * %-escaping is not used in path-scheme locations.
 */
public interface StpLocation extends javax.wvcm.Location
{
    /**
     * Returns the StpProvider object that created this StpLocation object
     * 
     * @return The StpProvider object that instantiated this instance of
     *         StpLocation
     */
    StpProvider stpProvider();

    /**
     * This class enumerates the namespaces that may appear in a location
     * specification. Instances of the class are used to represent the namespace
     * of the location represented by an StpLocation object.
     * 
     * The Namespace maps directly to the word token used in a location
     * specification to denote the namespace of the resource named by the
     * location. Each resource may appear in multiple namespaces.
     */
    public static enum Namespace implements StpExEnumeration, Serializable
    {
        /**
         * A special path-scheme Namespace indicating that the namespace field
         * of a location is unknown.
         */
        @RpcEnum(id=23, simpleName="INVALID")
        INVALID("INVALID" /* NOI18N */),

        /**
         * A special path-scheme Namespace indicating that no namespace was
         * specified in the location specification.
         */
        @RpcEnum(id=26, simpleName="NONE")
        NONE("NONE" /* NOI18N */),

        /**
         * A special path-scheme Namespace indicating that the namespace field
         * in the location specification was empty, which is the convention for
         * specifying the default namespace of a repository.
         */
        @RpcEnum(id=8, simpleName="DEFAULT")
        DEFAULT("DEFAULT" /* NOI18N */),

        /**
         * The special, compound namespace used in stable selector schemes. To
         * fully specify the stable selector scheme namespace, a resource-type
         * string must follow the REPO word token in the location specification.
         */
        @RpcEnum(id=38, simpleName="repo")
        REPO("repo" /* NOI18N */),

        /**
         * The special, compound namespace used in the selector scheme for
         * specification of efficiently-accessed locations. To fully specify the
         * efficient selector scheme namespace, a resource-type string must
         * follow the FAST word token in the location specification.
         */
        @RpcEnum(id=12, simpleName="fast")
        FAST("fast" /* NOI18N */),

        /**
         * A special file-path-scheme namespace that forces the rest of the
         * location to be interpreted as a ClearCase P-name, a file system
         * pathname with an optional history-mode extension.
         */
        @RpcEnum(id=28, simpleName="pname")
        PNAME("pname" /* NOI18N */),

        /**
         * A special file-path-scheme Namespace indicating a pname without an
         * explicit PNAME prefix. Locations in the PNAME_IMPLIED namespace
         * display simply as P-names.
         */
        @RpcEnum(id=29, simpleName="implicit")
        PNAME_IMPLIED("implicit" /* NOI18N */),

        /**
         * A repository-path-scheme Namespace for a ClearCase VOB specified
         * directly by tag or indirectly by an entity within the VOB.
         */
        @RpcEnum(id=46, simpleName="vob")
        VOB("vob" /* NOI18N */),

        /**
         * A repository-path-scheme Namespace for a ClearCase VOB specified by
         * its replica UUID
         */
        @RpcEnum(id=37, simpleName="replicauuid")
        REPLICA_UUID("replicauuid" /* NOI18N */),

        /**
         * A repository-path-scheme Namespace for a ClearCase VOB specified by
         * its VOB family UUID.
         */
        @RpcEnum(id=51, simpleName="vobuuid")
        VOB_UUID("vobuuid" /* NOI18N */),

        /**
         * A repository-path-scheme Namespace for a ClearCase view specified by
         * its UUID
         */
        @RpcEnum(id=45, simpleName="viewuuid")
        VIEW_UUID("viewuuid" /* NOI18N */),

        /**
         * A stable-selector-scheme Namespace for a ClearCase resource specified
         * by its DBID
         */
        @RpcEnum(id=6, simpleName="dbid")
        DBID("dbid" /* NOI18N */),

        /** A repository-path-scheme Namespace for a ClearQuest user database */
        @RpcEnum(id=43, simpleName="userdb")
        USER_DB("userdb" /* NOI18N */),

        /**
         * A repository-path-scheme Namespace for a ClearQuest database known
         * variously as a profile, connection, database-set, master database, or
         * schema repository
         */
        @RpcEnum(id=7, simpleName="dbset")
        DB_SET("dbset" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for an action
         */
        @RpcEnum(id=0, simpleName="action")
        ACTION("action" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for an activity
         */
        @RpcEnum(id=1, simpleName="activity")
        ACTIVITY("activity" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for an attribute type
         */
        @RpcEnum(id=2, simpleName="attype")
        ATTYPE("attype" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a baseline */
        @RpcEnum(id=3, simpleName="baseline")
        BASELINE("baseline" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a branch */
        @RpcEnum(id=50, simpleName="branch")
        BRANCH("branch" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a branch type */
        @RpcEnum(id=4, simpleName="brtype")
        BRTYPE("brtype" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a component */
        @RpcEnum(id=5, simpleName="component")
        COMPONENT("component" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for a dynamic choice
         * list.
         */
        @RpcEnum(id=10, simpleName="choicelist")
        DYNAMIC_CHOICE_LIST("choicelist" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for an element type */
        @RpcEnum(id=11, simpleName="eltype")
        ELTYPE("eltype" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a field definition */
        @RpcEnum(id=13, simpleName="field")
        FIELD_DEFINITION("field" /* NOI18N */),

        /** The URL-path-scheme and file-path-scheme Namespace for a file URL */
        @RpcEnum(id=14, simpleName="file")
        FILE("file" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a folder */
        @RpcEnum(id=15, simpleName="folder")
        FOLDER("folder" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a form */
        @RpcEnum(id=16, simpleName="form")
        FORM("form" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a group */
        @RpcEnum(id=17, simpleName="group")
        GROUP("group" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a hyperlink */
        @RpcEnum(id=18, simpleName="hlink")
        HLINK("hlink" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a hyperlink type */
        @RpcEnum(id=19, simpleName="hltype")
        HLTYPE("hltype" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a hook */
        @RpcEnum(id=20, simpleName="hook")
        HOOK("hook" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a label type */
        @RpcEnum(id=24, simpleName="lbtype")
        LBTYPE("lbtype" /* NOI18N */),

        /**
         * A stable-selector-scheme Namespace for a ClearCase resource specified
         * by its OID
         */
        @RpcEnum(id=27, simpleName="oid")
        OID("oid" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a pool */
        @RpcEnum(id=30, simpleName="pool")
        POOL("pool" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a project. */
        @RpcEnum(id=31, simpleName="project")
        PROJECT("project" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for a project
         * configuration
         */
        @RpcEnum(id=32, simpleName="projconfig")
        PROJECT_CONFIGURATION("projconfig" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for a query, chart,
         * report, report format, or query folder.
         */
        @RpcEnum(id=33, simpleName="query")
        QUERY("query" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for a record, record
         * type, attachment folder, or attachment
         */
        @RpcEnum(id=34, simpleName="record")
        RECORD("record" /* NOI18N */),

        /**
         * The user-friendly, efficient, and stable-scheme Namespace for a
         * ClearCase registry region.
         */
        @RpcEnum(id=35, simpleName="registryregion")
        REGISTRY_REGION("registryregion" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a replica */
        @RpcEnum(id=36, simpleName="replica")
        REPLICA("replica" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for an rptype */
        @RpcEnum(id=39, simpleName="rptype")
        RPTYPE("rptype" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a stream */
        @RpcEnum(id=40, simpleName="stream")
        STREAM("stream" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a trigger type */
        @RpcEnum(id=41, simpleName="trtype")
        TRTYPE("trtype" /* NOI18N */),

        /**
         * The URL-path-scheme Namespace denoting a location presented in the
         * form of an HTTP URI or URL
         */
        @RpcEnum(id=21, simpleName="http")
        HTTP("http" /* NOI18N */),

        /**
         * The URL-path-scheme Namespace denoting a location presented in the
         * form of an HTTPS URI or URL
         */
        @RpcEnum(id=22, simpleName="https")
        HTTPS("https" /* NOI18N */),

        /** The user-friendly-selector-scheme Namespace for a user */
        @RpcEnum(id=42, simpleName="user")
        USER("user" /* NOI18N */),

        /**
         * The user-friendly, efficient, and stable-scheme Namespace for a view
         * tag.
         */
        @RpcEnum(id=44, simpleName="viewtag")
        VIEWTAG("viewtag" /* NOI18N */),

        /**
         * The user-friendly, efficient, and stable-scheme Namespace for a VOB
         * tag.
         */
        @RpcEnum(id=47, simpleName="vobtag")
        VOBTAG("vobtag" /* NOI18N */),

        /**
         * The path-scheme Namespace for a ClearCase VOB tag, treated in some
         * senses as a directory.
         * 
         * Note: This namespace is an implementation detail of the server, which
         * unfortunately must be exposed to clients.
         * 
         * It is not intended that clients should attempt to create resources
         * within this namespace.  Undefined behavior is guaranteed if clients
         * attempt this.
         * 
         * It is not expected that the server will return resources within this
         * namespace.
         */
        @RpcEnum(id=48, simpleName="vobtagasdirectory")
        VOB_TAG_AS_DIRECTORY("vobtagasdirectory" /* NOI18N */),
        
        /** The path-scheme Namespace for a workspace */
        @RpcEnum(id=49, simpleName="workspace")
        WORKSPACE("workspace" /* NOI18N */),

        /**
         * The user-friendly-selector-scheme Namespace for a domain server;
         * e.g. ClearCase or ClearQuest
         */
        @RpcEnum(id=9, simpleName="domainserver")
        DOMAINSERVER("domainserver" /* NOI18N */),
        
        /** The path-scheme for a CCRC (aka, "legacy") server */
        @RpcEnum(id=25, simpleName="legacyserver")
        LEGACYSERVER("legacyserver" /* NOI18N */);

        /**
         * Returns the word token for this Namespace in the namespace field of a
         * location specification.
         * 
         * @return A String containing the namespace field value that denotes
         *         this Namespace.
         */
        public String toNamespaceField()
        {
            return m_namespaceField;
        }

        /**
         * Finds the Namespace enumerator from the identifier used in the
         * namespace field of a location specification.
         * 
         * @param field The word token as it appears in a location namespace
         *            field.
         * 
         * @return The Namespace that the namespace field identifier denotes.
         *         Namespace.INVALID is returned if no Namespace enumerator
         *         matches the symbol exactly.
         */
        public static final Namespace fromNamespaceField(String field)
        {
            // Force to lower case before consulting symbol map
            String symbol = field.toLowerCase();

            Namespace namespace = g_symbolToNamespaceMap.get(symbol);

            return namespace == null? Namespace.INVALID : namespace;
        }

        /**
         * EnumSet definitions for the various classifications of Namespaces
         */
        private static EnumSet<Namespace> invalid = EnumSet.of(INVALID, NONE);
        private static EnumSet<Namespace> valid = EnumSet.complementOf(invalid);
        private static EnumSet<Namespace> filePathSchemes =
            EnumSet.of(PNAME,
                       PNAME_IMPLIED,
                       FILE);
        private static EnumSet<Namespace> repositoryPathSchemes =
            EnumSet.of(VOB,
                       VIEW_UUID,
                       REPLICA_UUID,
                       USER_DB,
                       DB_SET,
                       LEGACYSERVER);
        private static EnumSet<Namespace> urlPathSchemes = EnumSet.of(FILE,
                                                                      HTTP,
                                                                      HTTPS);
        private static EnumSet<Namespace> pathSchemes =
            union(EnumSet.of(WORKSPACE, NONE, DEFAULT, INVALID),
                  filePathSchemes,
                  urlPathSchemes,
                  repositoryPathSchemes);
        private static EnumSet<Namespace> escapeEncoded =
            union(EnumSet.complementOf(union(pathSchemes, invalid)),
                  EnumSet.of(USER_DB, DB_SET));
        private static EnumSet<Namespace> extendedNamespaces =
            EnumSet.of(REPO, FAST);

        /**
         * Computes the union of a set of EnumSets
         * 
         * @param set Two or more EnumSet<Namespace> objects to be combined
         *            into one.
         * @return The logical union of the given EnumSets.
         */
        private static EnumSet<Namespace> union(EnumSet... set)
        {
            EnumSet<Namespace> result = EnumSet.noneOf(Namespace.class);

            for (EnumSet s : set)
                result.addAll(StpException.<EnumSet<Namespace>>unchecked_cast(s));

            return result;
        }
        
        /**
         * Determines whether this namespace is valid (not NONE or INVALID).
         * 
         * @return <b>true</b> if this Namespace represents a valid namespace;
         * <b>false</b> otherwise.
         */
        public boolean isValid()
        {
            return valid.contains(this);
        }

        /**
         * @return <b>true</b> iff this namespace prefixes a path scheme
         */
        public boolean isPathScheme()
        {
            return pathSchemes.contains(this);
        }

        /**
         * @return <b>true</b> iff this namespace prefixes a file path scheme.
         */
        public boolean isFilePathScheme()
        {
            return filePathSchemes.contains(this);
        }


        /**
         * @return <b>true</b> iff this namespace prefixes a path scheme
         *         selector for the name of a repository (or repository-like
         *         entity not in a repository). Said path is the value of the
         *         Repo field rather than the Name field of an StpLocation.
         */
        public boolean isRepositoryPathScheme()
        {
            return repositoryPathSchemes.contains(this);
        }


        /**
         * @return <b>true</b> iff this namespace prefixes a path scheme
         *         selector expressed as a URL or URI. The complete URI,
         *         including this prefix is contained wholly within the Name
         *         field of the StpLocation
         */
        public boolean isUrlPathScheme()
        {
            return urlPathSchemes.contains(this);
        }

        /**
         * @return <b>true</b> iff this namespace requires additional segments
         *         to complete its specification.
         */
        public boolean isExtendedNamespace()
        {
            return extendedNamespaces.contains(this);
        }

        /**
         * @return Whether or not the name field of a location specification
         *         prefixed by this namespace should be %-escaped encoded.
         */
        public boolean isEscapeEncoded()
        {
            return escapeEncoded.contains(this);
        }

        /**
         * Creates a new Namespace object given its namespace field image.
         * 
         * @param symbol The identifier used in the namespace field to denote
         *            this namespace.
         */
        private Namespace(String symbol)
        {
            m_namespaceField = symbol;
        }

        /** A map from namespace Symbol to Namespace */
        private static Map<String, Namespace> g_symbolToNamespaceMap =
            new Hashtable<String, Namespace>();

        static {
            for (Namespace n : Namespace.values())
                if (null != g_symbolToNamespaceMap.put(n.m_namespaceField,
                                                       n)) {
                    throw new IllegalArgumentException
                    ("Duplicate Selector.Namespace symbol" /* NOI18N */);
                }
        }

        /** The symbol used in a selector to denote this Namespace */
        private String m_namespaceField;

        /** Serialization version UID */
        private static final long serialVersionUID = -3736971155548723312L;
    }

    /**
     * The characters within a location specification that syntactically delimit
     * the fields of the selector.
     */
    static final String FIELD_DELIMITERS = "@" /* NOI18N */;

    /**
     * The characters within a selector field that syntactically delimit the
     * segments of a field.
     */
    static final String SEGMENT_DELIMITERS = "/\\" /* NOI18N */;

    /**
     * The characters within a selector that syntactically delimit the fields
     * and segments embedded within the selector.
     */
    static final String DELIMITERS = FIELD_DELIMITERS + SEGMENT_DELIMITERS;

    /**
     * If one of the characters of DELIMITERS is to be part of a name segment it
     * must be protected from its syntactic interpretation by preceding it with
     * this escape character. The escape character must also be escaped if it is
     * to be part of a name segment.
     */
    static final String ESCAPE_CHAR = "%" /* NOI18N */;

    /**
     * Overall status of this StpLocation
     * 
     * @return <b>true</b> if all required fields were found in the given
     *         location specification.
     */
    boolean isOk();

    /**
     * Generates an StpException object that reports the state of this
     * StpLocation.
     * 
     * @return An StpException whose message reports the state of this
     *         StpLocation. Will be <b>null</b> if this StpLocation is valid.
     */
    StpException status();

    /**
     * Throws an INVALID_OBJECT_SELECTOR StpException if this StpLocation does
     * not reflect a syntactically complete and correct location specification.
     * 
     * @throws StpException if any required fields are missing from the
     *             StpLocation specification.
     */
    void throwIfNotOk() throws StpException;

    /**
     * Returns whether or not this location is specified using a pathname
     * format. Such locations specify a location as a segmented pathname
     * following an explicit scheme prefix:
     * <p>[<i>domain</i> .] <i>namespace</i> : <i>segmented-path</i>.
     * <p>
     * The segmented-path is the value of either the name field or the repo
     * field of this StpLocation and the other field is not used and empty. The
     * segmented path is stored in the name field unless the predicate
     * {@link #isRepositoryPathScheme} is also <b>true</b>.
     * <p>
     * Included in this scheme classification are the location specifications
     * that are not complete enough to classify more precisely; i.e. it includes
     * the locations with the following special Namespace values.
     * <ul>
     * <li>NONE: The location specification did not include a scheme delimiter
     * (":") (at least, not before the first occurrence of a character not
     * allowed in a scheme prefix). The original input is in the name field of
     * this StpLocation.
     * <li>INVALID: The location specification began with a syntactically valid
     * scheme prefix, but the spelling of the namespace subfield did not match
     * any known namespace. The location was parsed as a path-scheme location
     * with all text following the first ':' stored in the name field. The
     * apparently misspelled namespace field is available from the
     * ExtendedNamespace field of this StpLocation.
     * <li>DEFAULT: The location specification began with a syntactically valid
     * scheme prefix, but the namespace field was empty. The location was parsed
     * as a path-scheme location, with all text following the first ':' stored
     * in the name field.
     * </ul>
     * 
     * @return <b>true</b> if this is a path-scheme location; <b>false</b> otherwise
     */
    boolean isPathScheme();

    /**
     * Returns whether or not this location is specified using the URL path-
     * scheme format. Such locations are formatted as standards-conforming URLs
     * (URL-encoded). The entire URL, including the scheme-prefix of the URL,
     * such as "http:" or "file:" is included in the name field of the
     * StpLocation object. An optional domain field is permitted before the
     * scheme-prefix, but it is not included in the name field. A URL path
     * scheme is a specialized form of path scheme. The URL could designate a
     * server, a repository, or a resource inside or outside of a repository.
     * The URL is stored in the repo field if {@link #isRepositoryPathScheme} is
     * <b>true</b>; otherwise it is stored in the name field.
     * 
     * @return <b>true</b> if this is a URL path-location; <b>false</b>
     *         otherwise
     */
    boolean isUrlPathScheme();

    /**
     * Returns whether or not this location specifies a repository using a path-
     * scheme format. Such locations have a repository field specified as a
     * segmented pathname (un-encoded). A repository path-scheme is a
     * specialized form of path-scheme in which the path is found in the repo
     * field of the StpLocation object. For all other forms of path-location the
     * path is found in the name field of the StpLocation object.
     * 
     * @return <b>true</b> if this is a repository path-location; <b>false</b>
     *         otherwise
     */
    boolean isRepositoryPathScheme();

    /**
     * Returns whether or not this location is specified using the file path-
     * scheme format. In this format, the resource location is specified in the
     * name field of this StpLocation as a segmented pathname (using native file
     * system encoding conventions) to a file system object, perhaps extended by
     * a ClearCase history-mode selector. The variant of the file path location
     * format used in the specification of this StpLocation is indicated by the
     * value of getNamespace().
     * <ul>
     * <li>FILE: The "file:" URL scheme prefix was used to specify this
     * StpLocation. Since this is also a URL path-scheme, the file-scheme prefix
     * is included in the name field of this object. Conversion of this location
     * to a File via {@link #getFile()} or canonical path via
     * {@link #getCanonicalPath()} will first use the java.net.URI class to
     * parse the file URL.
     * <li>PNAME: This location was specified with an explicit "pname:" prefix.
     * The "pname:" prefix is <i>not</i> included in the name field of this
     * object; it contains only the characters following "pname:"
     * <li>PNAME_IMPLIED: This location was not specified with an explicit
     * "pname:" prefix but is being treated as if it were a pname. The implied
     * "pname:" prefix is <i>not</i> included in the name field of this object
     * nor does it appear in the string image of this StpLocation.
     * </ul>
     * <p>
     * Note that this and the other predicates are purely syntactic. The user
     * may have intended to name a file, but if it so happens that its name
     * looks exactly like a valid object selector, it will be parsed and
     * classified as an object selector. {@link #isObjectSelectorScheme()} will
     * be <b>true</b> not {@link #isFilePathScheme()}. Clients wishing to
     * interpret a location as a file path location, may always use the
     * {@link #getFile()} or {@link #getCanonicalPath()} methods to investigate
     * that option further. If this StpLocation isn't in the FILE or PNAME
     * namespace, these methods will use the original input in its entirety as
     * the intended pathname.
     * <p>
     * Similarly, {@link #recomposeWithNamespace(StpLocation.Namespace) 
     * recomposeWithNamespace(Namespace.PNAME)} will "do the right thing" and
     * force the original input into an explicit file path selector. Note,
     * however, that in this case, the image of that StpLocation will include
     * the "pname:" prefix.
     * 
     * @return <code><b>true</b></code> if this selector is most likely a
     *         pathname to a file system object, <code><b>false</b></code> if
     *         there is a more likely interpretation.
     */
    boolean isFilePathScheme();

    /**
     * Returns whether or not this file-path scheme location uses the optional
     * ClearCase history-mode naming syntax.
     * 
     * @return <b>true</b> if the name segment of this location contains
     *         history-mode naming syntax.
     */
    boolean isHistoryModeScheme();

    /**
     * Returns whether or not this location uses either a stable, fast
     * (efficient), or user-friendly object selector scheme. Locations using the
     * object selector format have a pre-defined namespace and separate name and
     * repository fields.
     * 
     * @return <b>true</b> if this location uses the object selector format;
     *         <b>false</b> otherwise, in which case it uses either a path
     *         scheme.
     */
    boolean isObjectSelectorScheme();

    /**
     * Returns whether or not this location uses an object selector scheme with
     * user-friendly namespace, name, and repository fields.
     * 
     * @return <b>true</b> if this is an object name selector; <b>false</b>
     *         otherwise.
     * 
     */
    boolean isUserFriendlySelectorScheme();

    /**
     * Returns whether or not this location uses an object selector scheme with
     * a compound REPO namespace. Its name and repository fields are densely
     * encoded for greater stability and more efficient retrieval.
     * 
     * @return <b>true</b> if the location uses the REPO namespace; <b> false</b>
     *         otherwise.
     */
    boolean isRepoSelectorScheme();

    /**
     * Returns whether or not this location uses an object selector scheme with
     * a compound FAST namespace. Its name and repository fields are densely
     * encoded for greater stability and more efficient retrieval.
     * 
     * @return <b>true</b> if the location uses the FAST namespace; <b> false</b>
     *         otherwise.
     */
    boolean isFastSelectorScheme();

    /**
     * Returns whether or not this location uses an object selector scheme with
     * a compound OID namespace. Its name and repository fields are densely
     * encoded for greater stability and more efficient retrieval.
     * 
     * @return <b>true</b> if the location uses the OID namespace; <b> false</b>
     *         otherwise.
     */
    boolean isOidSelectorScheme();

    /**
     * Returns the StpLocation.Namespace of this selector.
     * <p>
     * The special Namespace.INVALID indicates that the namespace field was
     * present but spelled different from any namespace known to the library.
     * <p>
     * The special Namespace.DEFAULT indicates that the namespace field was
     * present but empty, indicating that the default namespace ought to be
     * used.
     * <p>
     * The special Namespace.NONE indicates that the namespace field was not
     * present (i.e. there was no ':' in the specification before the first
     * occurrence of a character not allowed in a scheme prefix), making it
     * quite likely that this is a file selector.
     * <p>
     * Namespace.HTTP, Namespace.HTTPS, and Namespace.FILE indicate that the
     * selector used the URI/URL syntax, the entirety of which is present in the
     * name property.
     * <p>
     * Namespace.PNAME indicates that the selector used the PNAME namespace
     * prefix. The file pathname following the PNAME prefix is the value of the
     * name field.
     * <p>
     * See the complete list of possible namespaces in the Namespace enum
     * specification.
     * 
     * @return The namespace used in this location specification as a Namespace
     *         object. This will never be <b>null</b>.
     */
    Namespace getNamespace();

    /**
     * Returns the resource type field of a location specification if it used a
     * compound namespace.
     * 
     * @return The resource type segment of this StpLocation. This field is
     *         defined only for compound namespace locations (i.e. those that
     *         use Namespace.REPO, Namespace.FAST, or Namespace.OID). It will be
     *         an empty string otherwise.
     */
    String getResourceType();

    /**
     * An object containing additional information associated with certain
     * Namespace enumerators.
     */
    interface ExtendedNamespace
    {

        /**
         * @return Returns the namespace.
         */
        Namespace getNamespace();

        /**
         * @return Returns the resource type portion of the compound REPO
         *         namespace field. Returns the empty string if the namespace is
         *         not REPO, FAST, or OID.
         */
        String getResourceType();

        /**
         * The image of the extended namespace field. For Namespace.REPO, FAST,
         * or OID, this includes the resource type subfield; for
         * Namespace.INVALID, this is the (misspelled) namespace field as
         * entered.
         * 
         * @return A String containing the image of the namespace field for this
         *         extended namespace (without the delimiting ':'). Will be
         *         empty if the namespace is DEFAULT and <b>null</b> if the
         *         location specification has no namespace field separate from
         *         the name field, i.e. if the namespace is NONE, PNAME_IMPLIED,
         *         HTTP, HTTPS, or FILE.
         */
        String toNamespaceField();

        /**
         * Generates a debug image for this ExtendedNamespace object
         * 
         * @return Returns the value of {@link #toNamespaceField()},
         *         substituting "&lt;null>" for <b>null</b>.
         */
        public String toString();
    }

    /**
     * Returns an ExtendedNamespace object that, for some namespaces, contains
     * additional information about the namespace field beyond its Namespace
     * value.
     * 
     * @return For a REPO, FAST, or OID namespace, the ExtendedNamespace
     *         specifies the resource type segment that is associated with it;
     *         for an INVALID namespace, the ExtendedNamespace object specifies
     *         the misspelled namespace field; for other namespaces, no
     *         additional information is available. Will not be <b>null</b>.
     */
    ExtendedNamespace getExtendedNamespace();

    /**
     * Returns the object name field specified for this location. This field is
     * relevant and meaningful in all schemes <i>except</i> the repository-path
     * scheme. In a repository-path scheme, it will be an empty Sting. The
     * encoding of the returned String is unchanged from the original input.
     * 
     * @return An empty string for a repository-path-scheme location; otherwise
     *         the object name field of an object selector or the pathname of a
     *         path-scheme location. Will be never be <b> null</b>, but may be
     *         empty.
     */
    String getName();

    /**
     * The number of segments in the object name.
     * 
     * @return The length of the Sting array returned by
     *         {@link #getNameSegments(int) getNameSegments(Integer.MAX_VALUE)}.
     */
    int getNameSegmentCount();

    /**
     * Returns the first N segments of the name field of this location
     * specification. If the requested number of segments is greater than the
     * number in the name field, the entire name is returned; if zero or less,
     * an empty array is returned.
     * <p>
     * Constructs a String array containing the segments of the given field. The
     * elements of the array are the character sequences that preceded each
     * segment delimiter plus the character sequence at the end of the field not
     * followed by a delimiter as long as it is not empty. Thus, the array is
     * empty if the field is empty; otherwise, the array has N+1 segments, where
     * N is the number of segment delimiters in the field not counting the last
     * delimiter if it appears at the end of the field.
     * <p>
     * The following examples illustrate the way a field is segmented.
     * 
     * <pre>
     *  "" ==> {}
     *  "fob/bar" ==> {"fob", "bar"}
     *  "fob/" ==> {"fob"}
     *  "fob" ==> {"fob"}
     *  "/fob" ==> {"", "fob"}
     *  "/" ==> {""}
     *  "//" ==> {"", ""}
     *  "fob//bar" ==> {"fob", "", "bar"}
     *  "http://server" ==> {"http:", "", "server"}
     *  "http://" ==> {"http:", ""}
     *  "file:///" ==> {"file", "", ""}
     * </pre>
     * 
     * Note that a trailing segment delimiter is "lost" only if it follows a
     * non-empty segment. Consequently, when reconstructing the field from the
     * array, segment delimiters should be inserted between each array element
     * and a trailing delimiter should be added only if the last segment is
     * empty.
     * 
     * <p>
     * Note: The returned segments are encoded just as they were on input to the
     * constructor. Any escape characters present in the field on input remain
     * in each returned segment. Only the unescaped segment delimiters have been
     * removed from the input field.
     * 
     * @param nSegs the number of segments to return
     * 
     * @return a String array containing the first nSegs segments of the name
     *         field of this StpLocation.
     */
    String[] getNameSegments(int nSegs);

    /**
     * Returns contiguous segments of the name field of this selector. Segments
     * returned are in the intersection of the specified range and the actual
     * range of name segments. The first segment is at index zero.
     * 
     * <p>
     * Note: The returned segments are encoded just as they were on input to the
     * constructor. Any escape characters present in the field on input remain
     * in each returned segment. Only the unescaped segment delimiters have been
     * removed from the input field.
     * 
     * @param firstSeg the first segment to include
     * @param lastSeg the last segment to include
     * 
     * @return the requested segments of the object name. Will never be null,
     *         but may be empty if the specified range includes none of the
     *         segments of the name field.
     */
    String[] getNameSegments(int firstSeg,
                             int lastSeg);

    /**
     * Returns the repository field of this location specification. This field
     * is irrelevant and empty in any path scheme location that does not have a
     * repository field. Conversely, if this location specification specified a
     * repository, it will be in this field.
     * 
     * @return An empty string if there was no repository field found in the
     *         location specification; otherwise the image of the repository
     *         field (without a repository field delimiter).
     */
    String getRepo();

    /**
     * The number of segments in the repository name.
     * 
     * @return The length of the String array returned by
     *         {@link #getRepoSegments(int) getRepoSegments(Integer.MAX_VALUE)}.
     */
    int getRepoSegmentCount();

    /**
     * Returns the first N segments of the repository field of this location. If
     * the requested number of segments is greater than the number in the
     * repository name, the entire repository name is returned; if zero or less,
     * an empty array is returned.
     * 
     * <p>
     * Note: The returned segments are encoded just as they were on input to the
     * constructor. Any escape characters present in the field on input remain
     * in each returned segment. Only the unescaped segment delimiters have been
     * removed from the input field.
     * 
     * @param nSegs the number of segments to return
     * 
     * @return a String containing the first nSegs segments of the repository
     *         name.
     * 
     * @see #getNameSegments(int) for a description of how segments are parsed
     *      and counted.
     */
    String[] getRepoSegments(int nSegs);

    /**
     * Returns contiguous segments of the repository name of this location.
     * Segments returned are in the intersection of the specified range and the
     * actual range of name segments. The first segment is at index zero.
     * 
     * <p>
     * Note: The returned segments are encoded just as they were on input to the
     * constructor. Any escape characters present in the field on input remain
     * in each returned segment. Only the unescaped segment delimiters have been
     * removed from the input field.
     * 
     * @param firstSeg the first segment to include
     * @param lastSeg the last segment to include
     * 
     * @return the requested segments of the repository name. Will never be
     *         null, but may be empty if the specified range includes none of
     *         the segments of the name.
     */
    String[] getRepoSegments(int firstSeg,
                             int lastSeg);

    /**
     * Returns the domain specified or implied by the selector. In
     * URL-path-scheme locations this field is optional and will be NONE if no
     * domain information is available. In the other formats, the value NONE
     * denotes the default domain.
     * 
     * @return Returns the StpProvider.Domain. Will never be <b>null</b>, but
     *         may be {@link StpProvider.Domain#NONE NONE} or {@link
     *         StpProvider.Domain#INVALID INVALID}.
     */
    StpProvider.Domain getDomain();

    /**
     * Reconstitutes the location specification from its component fields.
     * 
     * @return For a valid StpLocation, a syntactically correct location
     *         specification string composed from the current values for the
     *         namespace, name, domain, and repo fields; otherwise the location
     *         specification as passed to the constructor.
     * 
     * @see java.lang.Object#toString()
     */
    String toString();

    /**
     * As above, but returns a location string <i>without</i> the domain
     * prefix.
     */
    String toStringWithoutDomain();

    /**
     * @see java.lang.Object#equals(java.lang.Object)
     */
    boolean equals(Object arg0);

    /**
     * Uses the hash code of the composed String image
     * 
     * @see java.lang.Object#hashCode()
     */
    int hashCode();

    /**
     * Constructs an StpLocation object based on the fields of this StpLocation
     * with optional replacements for some of the fields. A <b>null</b>
     * argument generally means to use the corresponding field of this
     * StpLocation in the new StpLocation.
     * <p>
     * NOTE: This method does not change the host StpLocation object. But
     * constructs and returns a new instance of StpLocation.
     * 
     * @param namespace The namespace for the new location expressed either as
     *            an ExtendedNamespace object, a Namespace enumeration or as a
     *            String containing the resource type of a REPO namespace. If
     *            namespace is Namespace.NONE no namespace prefix is generated
     *            for the selector. If namespace is <b>null</b>, the current
     *            value of getExtendedNamespace() is used.
     * @param name The name field of the new selector. If <b>null</b>, the
     *            current value of getName() is used.
     * @param domain The StpProvider.Domain for the new selector. If null, the
     *            current value of getDomain() is used.
     * @param repo The repository field for the new selector. If <b>null</b>,
     *            the current value of getRepo() is used. Must be <b>null</b>
     *            for path-path scheme locations that are not
     *            repository-path-schemes. If empty, no repository field will be
     *            generated for the selector.
     * 
     * @return A new StpLocation composed from the current namespace, name, type
     *         and repo fields, optionally overwritten by the given arguments.
     * 
     * @throws StpException Thrown if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithMods(Object namespace,
                                  String name,
                                  StpProvider.Domain domain,
                                  String repo) throws StpException;

    /**
     * Constructs new location based on this location but with a replacement for
     * its namespace field.
     * 
     * @param namespace The namespace for the new StpLocation. If namespace is
     *            Namespace.NONE, no namespace prefix is generated for the
     *            selector. If namespace is <b>null</b>, the current value of
     *            Namespace is used, effectively cloning this StpLocation
     *            object.
     * 
     * @return An StpLocation composed from the current name, domain and repo
     *         fields and the specified namespace argument.
     * 
     * @throws StpException if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithNamespace(StpLocation.Namespace namespace)
        throws StpException;

    /**
     * Constructs a new location based on this location with a replacement for
     * its name field.
     * 
     * @param name The new selector name field. If null, the current value of
     *            name() is used.
     * 
     * @return The selector composed from the current namespace, domain and repo
     *         fields and the specified name argument.
     * 
     * @throws StpException Thrown if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithName(String name) throws StpException;

    /**
     * Constructs a new location based on this location with a replacement for
     * its repository field.
     * 
     * @param repo The new repository field for the location. If <b>null</b>,
     *            the current value of getRepo() is used. If empty, no
     *            repository field will be generated for the location.
     * 
     * @return An StpLocation composed from the current namespace, name, and
     *         domain fields and the specified repo argument.
     * 
     * @throws StpException Thrown if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithRepo(String repo) throws StpException;

    /**
     * Constructs a new location based on this location with a replacement for
     * its domain field.
     * 
     * @param domain The new domain for the selector. If <b>null</b>, the
     *            current value of getDomain() is used.
     * 
     * @return An StpLocation composed from the current namespace, name, and
     *         repo fields and the specified domain argument.
     * 
     * @throws StpException Thrown if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithDomain(StpProvider.Domain domain)
        throws StpException;

    /**
     * Constructs an object selector with a replacement for its resource type
     * field, forcing the namespace to REPO.
     * 
     * @param rType The resource type for the new location. If <b>null</b>, the
     *            current value of getResourceType() is used.
     * 
     * @return A stable-selector scheme StpLocation composed from the current
     *         name, repo and domain fields and the specified resource type.
     * 
     * @throws StpException Thrown if the given selector String is not in the
     *             correct form. StpReasonCode=INVALID_OBJECT_SELECTOR
     */
    StpLocation recomposeWithResourceType(String rType) throws StpException;

    /**
     * Constructs an StpLocation for a pname based on the image of this
     * StpLocation. The StpLocation can be constructed with or without a pname
     * prefix and can be assigned a StpProvider.Domain.
     * <p>
     * In most cases, the entire image of this StpLocation (as returned by
     * toString()) becomes the value of the name field of the returned
     * StpLocation even if that image includes a namespace and/or a domain
     * prefix. The domain prefix can be elided from the name field of the new
     * location by using <b>null</b> for the domain argument to this method.
     * <p>
     * To convert any ill-formed selector to an implied pname, the following
     * logic might be used <code><pre>
     * if (!myLoc.isOk()) {
     *      // Convert to an implied pname so that it prints as it was
     *      // entered, but is treated internally as an OK file path selector
     *      myLoc = myLoc.recomposeAsPname(false, MY_DOMAIN);
     * }
     * </pre></code>
     * <p>
     * To convert any input not already formatted as a file-path-scheme location
     * to an explicit pname, the following logic might be used <code><pre>
     * if (!(myLoc.isFilePathScheme() && myLoc.isOk())) {
     *      myLoc = myLoc.recomposeAsPname(true, MY_DOMAIN);
     * }
     * </pre></code>
     * <p>
     * To convert all input to an implied pname, the following logic might be
     * used <code><pre>
     *  if (myLoc.getNamespace() == Namespace.PNAME) {
     *      // Keep the original "pname:" prefix out of the implied pname
     *      // Preserve any domain data from the input. 
     *      myLoc = myLoc.recomposeWithNamespace(Namespace.PNAME_IMPLIED);
     *  } else if (myLoc.getNamespace() == Namespace.FILE) {
     *      // Remove any domain info from the "file:" prefix, but push the
     *      // rest of the "file:" prefix into the pname.
     *      myLoc = myLoc.recomposeAsPname(false, null);
     *  } else {
     *      // All other input not using an explicit pname: or file: prefix is
     *      // treated as a raw pname.
     *      myLoc = myLoc.recomposeAsPname(false, MY_DOMAIN);
     *  }
     *  
     *  // Set the domain type if not already specified.
     *  if (myLoc.getDomain() == Domain.NONE)
     *      myLoc = recomposeWithType(MY_DOMAIN);
     * </pre></code>
     * 
     * @param withPrefix if <b>true</b>, the namespace of the returned
     *            StpLocation will be Namespace.PNAME; if <b>false</b>, the
     *            namespace will be Namespace.PNAME_IMPLIED.
     * @param domain The StpProvider.Domain of the returned StpLocation. If
     *            <b>null</b> the StpProvider.Domain of the current StpLocation
     *            will be used <i>and the image of that domain will be elided
     *            from the pname.</i>
     * @return An StpLocation reconfigured as a Pname instance, explicit or
     *         implied as requested
     */
    StpLocation recomposeAsPname(boolean withPrefix,
                                 StpProvider.Domain domain) throws StpException;

    /**
     * Constructs a location suitable for addressing resources of the type
     * indicated by the supplied proxy class by filling in unspecified fields of
     * this location using provider-defined or resource-type-dependent defaults.
     * <p>
     * Note, when an StpLocation object is passed to a Provider proxy factory
     * method this forClass is implicitly invoked using the Class of the proxy
     * returned by the proxy factory method. Similarly, when an StpLocation is
     * passed to a method of a proxy, that method implicitly invokes forClass
     * using a proxy class deduced from the host proxy and the operation.
     * <p>
     * Clients need to use this method only if they want to complete/verify a
     * location used in some other context or more tightly than can be
     * determined from the proxy context.
     * 
     * @param proxyClass The Class object for the proxy interface for which this
     *            location is to be completed.
     * @return An StpLocation suitable for use with a proxy of the given class.
     *         The result will be this location if it is already suitable and a
     *         new StpLocation if not; will never be <b>null</b>.
     * @throws WvcmException if it is not possible to complete the location from
     *             available defaults or if the completed location is
     *             inappropriate in some other (obvious) way, such as having a
     *             domain or namespace inconsistent with the given class.
     */
    StpLocation forClass(Class<? extends Resource> proxyClass) throws WvcmException;

    /**
     * Returns an StpLocation whose segmented name field is one segment shorter
     * than the name field of this StpLocation <i>provided</i> the name field
     * of the resulting StpLocation would be valid. All other fields of the
     * StpLocation are left unchanged.
     * <p>
     * NOTE: For repository-path scheme locations this method operates on the
     * segmented repo field rather than the name field. But in all other
     * respects the behavior is the same.
     * <p>
     * For object-selector scheme locations, an empty name field is valid, but
     * for path-scheme locations, the name/repo field is valid only if it
     * contains at least one segment of the original path (even if that segment
     * is empty). These examples illustrate the edge cases <code><pre>
     *      StpLocation        |    Parent        
     *      -------------------+---------------
     *      /food              | /        
     *      vob:/food          | vob:/      
     *      http://server/path | http://server 
     *      file://author/path | file://author 
     *      http://server      | &lt;null&gt; 
     *      file://author      | &lt;null&gt; 
     *      pname:path         | &lt;null&gt;      
     *      file:/             | &lt;null&gt;      
     * </pre></code>
     * 
     * @return A new StpLocation instance; will be <b>null</b> if the name
     *         field of this StpLocation has no segments that can be removed.
     */
    Location parent();

    /**
     * Returns an StpLocation whose name field is the name field of this
     * StpLocation extended by the given child segment. All other fields are the
     * same as the fields of this StpLocation. For repository-path-scheme
     * locations, the repo field is extended rather than the name field.
     * <p>
     * Unlike most of the other methods of this class, the child() method
     * encodes the new child segment according to the requirements of the
     * scheme. Thus, this method may be used to add only one segment at a time
     * to the StpLocation. In any scheme, any embedded segment delimiters in the
     * child segment will be encoded to make them part of the segment.
     * <p>
     * Even if this method successfully returns an StpLocation, there is no
     * guarantee that the returned location is a valid resource location. The
     * returned location may be invalid even if the original location was valid.
     * Some resources simply do not have parents even though their location
     * suggests that they do.
     * <p>
     * For example, <b>field:Defect/Headline@7.0.0.0/SAMPL</b> is the location
     * for the description of the <b>Headline</b> field of the <b>Defect</b>
     * record type in the sample ClearQuest database. However, its parent
     * location, <b>field:Defect@7.0.0.0/SAMPL</b>, is <i>not</i> a valid
     * location. While this may seem to address the <b>Defect</b> record type
     * resource, it does not. The location for the <b>Defect</b> record type
     * resource is, in fact, <b><u>record</u>:Defect@7.0.0.0/SAMPL</b>, which
     * is in a different namespace from the parent of the field description
     * resource.
     * <p>
     * In general, clients are discouraged from manipulating locations to
     * traverse the object model. They should use the properties defined for
     * this purpose instead. If, for example, the client wants to traverse from
     * a field description to the record type of that field, then it should use
     * the RECORD_TYPE property of the field rather than taking the parent and
     * changing the namespace. Note that if the field location is a
     * stable-selector scheme location, simply changing the namespace of the
     * parent will not work.
     * 
     * @param child The new segment to be appended to the name field of this
     *            StpLocation. To be consistent with the Location.child method,
     *            it is assumed that the String is not yet encoded. It will be
     *            encoded as a single segment before adding it to the name
     *            field.
     * 
     * @return A new StpLocation with an extended name field.
     */
    Location child(String child);

    /**
     * Returns the last segment of the name field of this StpLocation. (Returns
     * the last segment of the repo field for repository-path-scheme locations.)
     * Any encoding used within the last segment is removed before returning a
     * value. Thus it's the case that
     * <b>loc.equals(loc.parent().child(loc.lastSegment()))</b> as long as
     * <b>loc.parent()</b> is not null.
     * <p>
     * At the root of the namespace (parent() returns &lt;null>), returns either
     * the name of the root or the empty string if the root is unnamed. These
     * examples illustrate the edge cases <code><pre>
     *      StpLocation        | lastSegment | parent        
     *      -------------------+-------------+------------- 
     *      /food              |   food      | /   
     *      /                  |  &lt;empty&gt;    | &lt;null&gt;       
     *      pname:/food        |   food      | pname:/
     *      http://server/path |   path      | http://server
     *      file://author/path |   path      | file://author
     *      http://server      |  &lt;empty&gt;    | &lt;null&gt;
     *      file://author      |  &lt;empty&gt;    | &lt;null&gt;
     *      pname:path         |   path      | &lt;null&gt;
     *      pname:/            |  &lt;empty&gt;    | &lt;null&gt;
     *      pname:\            |  &lt;empty&gt;    | &lt;null&gt;
     *      record:food@cq:s/u |   food      | record:@cq:s/u
     *      record:@cq:s/u     |  &lt;empty&gt;    | &lt;null&gt;
     * </pre></code>
     * 
     * @return A String containing the last segment of the name field of this
     *         StpLocation stripped of all encodings. Will never be null, but
     *         may be empty if the last segment is unnamed.
     */
    String lastSegment();

    /**
     * Interprets this StpLocation as a file-path-scheme location and returns
     * the <i>canonical</i> pathname for the file. If this StpLocation is not a
     * file-path-scheme location, the original location specification used to
     * construct this StpLocation is used as the pathname to be canonicalized.
     * <p>
     * For a location in the FILE namespace, this method constructs a URI from
     * the given file-scheme URL and then constructs a java.io.File from that
     * URI. Whether or not this succeeds depends on the JVM. (IBM's JVM 1.4.2,
     * for example, requires that the authority portion be empty.)
     * 
     * @see java.io.File#getCanonicalPath()
     * 
     * @return The canonicalized pathname for this resource. Will never be
     *         <b>null</b>.
     * 
     * @throws StpException if IO errors are encountered while determining the
     *             canonical path or converting the file-scheme URL to a File.
     */
    String getCanonicalPath() throws StpException;

    /**
     * Returns a File object that references the path defined by this
     * StpLocation. If this StpLocation is not a file-path-scheme location, the
     * original location specification used to construct this StpLocation is
     * used as the pathname from which the java.io.File object is constructed.
     * <p>
     * For a location in the FILE namespace, this method constructs a URI from
     * the given URL and then constructs a java.io.File from that URI. Whether
     * or not this succeeds depends on the JVM. (IBM's JVM 1.4.2, for example,
     * requires that the authority portion be undefined.)
     * 
     * @return A File object for the path defined by this StpLocation; Will
     *         never be <b>null</b>.
     * 
     * @throws MalformedURLException if the selector is a file scheme URL for
     *             which a File cannot be constructed.
     * @throws StpException
     * @throws IllegalStateException If {@link #isFilePathScheme()} is <b> false</b>.
     */
    File getFile() throws MalformedURLException, StpException;

}
